#!/usr/bin/perl

use warnings;
use strict;

my $class = 'xillyusb';
my $basedir = '/sys/bus/usb/drivers';

my @interfaces;
my %usbdevices;

foreach my $f (<"$basedir/$class/*">) {
  next unless ($f =~ /\d+-[^\/]*$/); # Skip anything but N-* file names
  my $l = readlink($f);

  unless (defined $l) {
    warn("Unexpected sysfs file $f\n");
    next;
  }
  push @interfaces, $l;
}

unless (@interfaces) {
  print STDERR "No matching device found.\n";
  print STDERR "Is the device connected? Is it recognized?\n";
  print STDERR "Is the XillyUSB driver loaded?\n";

  exit(1);
}

# The loop is on interfaces, even though what's listed is devices. Hence
# devices with multiple interfaces show repeated information. This makes no
# difference with XillyUSB, which presents a single interface.

foreach my $i (@interfaces) {
  my @path = split('/', "$basedir/$class/$i");

  pop @path; # The interface entry isn't interesting

  print "Connection of USB device: $path[-1]\n";
  print "===================================================\n";

  # Find the path of the host controller
  my $pos;

  for ($pos = 0; $pos < @path; $pos++) {
    last if ($path[$pos] =~ /^usb\d+$/);
  }

  die("Something is wrong with the path to the interface: $i\n")
    if ($pos == @path);

  my $controllerpath = join('/', @path[0..($pos-1)]);

  my $ueventinfo = readfile("$controllerpath/uevent");

  my ($pcivendor, $pciproduct) = ($ueventinfo =~ /^PCI_ID=([0-9a-fA-F]{4}):([0-9a-fA-F]{4})/mi);
  my ($driver) = ($ueventinfo =~ /^DRIVER=(.+?)[ \t\n\r]*$/mi);

  if (defined $pcivendor) {
    my $pcidev = pci_dev($pcivendor, $pciproduct);
    print "USB Host controller: PCI ID ".lc("$pcivendor:$pciproduct\n");
    print "$pcidev\n";
  } else {
    print "Non-PCI USB Host controller.\n";
  }
  print "Linux driver: $driver\n" if (defined $driver);

  print "\nPath from USB root hub to device:\n";

  for (; $pos < @path; $pos++) {
    $usbdevices{$path[$pos]} = 1;

    my $usbpath = join('/', @path[0..$pos]);
    my $what = ($pos < $#path) ? 'USB Hub element' : 'USB Device';

    print "\n--- $what: ".readfile("$usbpath/idVendor").':'.
      readfile("$usbpath/idProduct")."\n";
    print "    Manufacturer: ".readfile("$usbpath/manufacturer")."\n";
    print "    Product: ".readfile("$usbpath/product")."\n";
 }

  print "\n";
}

print "System information\n";
print "==================\n";

foreach my $f (qw[ bios_vendor bios_date
		   board_vendor board_name board_version
		   product_name product_version
		   sys_vendor ]) {

  print "$f: ".readfile("/sys/class/dmi/id/$f")."\n";
}

print "\nKernel version: ".readfile('/proc/version')."\n";

my $cpuinfo = readfile('/proc/cpuinfo');
my ($cpu) = ($cpuinfo =~ /model name[ \t:]+(.+)/);

print "CPU: $cpu\n";

# Run dmesg and fetch only lines containing the identifiers of the USB
# devices found: Prepeare a regular expression which is a (?:X|Y|Z)
# combination of all these. But the identifiers also need to be escaped,
# which is what the \Q and \E thing is for. Plus get all lines matching
# "xillyusb", case-insensitive. Some Perl regex wizardry.

my $logregex = join('|', qr/xillyusb/i, map { qr/\Qusb $_:\E/ } sort keys %usbdevices);

print "\n===================================================\n";
print "\nRelated lines from the kernel log:\n\n";

my $dmesg = `dmesg`;

my @loglines = grep /(?:$logregex)/, split(/[\r\n]+/, $dmesg);
splice @loglines, 0, -100; # Remove all but 100 last lines
print map { "$_\n" } @loglines; # Print at most 100 lines

exit(0);

################### SUBROUTINES ###################

sub readfile {
  my $fname = shift;
  local $/; # Slurp mode

  open (my $fh, "<", $fname) || return '(unavailable)';
  my $d = <$fh>;
  close $fh || return '(error)';

  $d =~ s/[ \t\n\r]+$//; # Remove possible trailing CR/LF and whitespace
  return $d;
}

sub pci_dev {
  my ($vendor, $product) = map { lc } @_;

  open (my $fh, "<", '/usr/share/misc/pci.ids') ||
    return "Device $vendor:$product (No pci.ids file found on this computer)";

  my ($l, $vstr, $pstr);

  while (defined ($l = <$fh>)) {
    ($vstr) = ($l =~ /^$vendor[ \t]+(.*?)[ \t\r\n]*$/);
    last if (defined ($vstr));
  }

  unless (defined $vstr) {
    return "Unknown device $vendor:$product";
    close $fh;
  }

  while (defined ($l = <$fh>)) {
    last if ($l =~ /^[0-9a-fA-F]{4}/); # Quit loop when reaching next vendor ID
    ($pstr) = ($l =~ /^\t$product[ \t]+(.*?)[ \t\r\n]*$/);
    last if (defined ($pstr));
  }

  close $fh;

  $pstr = "Device $product"
    unless (defined $pstr);

  return "$vstr: $pstr";
}
